package ua.in.hutorny.fsa.util;

import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;

import ua.in.hutorny.fsa.model.*;

public class Walker {
	final private WalkerCallback callback;  
	private Class<? extends State> rootClass = State.class;
	private Set<Class<? extends State>> processedClasses = new HashSet<Class<? extends State>>();
	private Set<Method> processedMethods = new HashSet<Method>();
	private Set<Class<? extends State>> disabledStates = new HashSet<Class<? extends State>>();
	
	public Walker(WalkerCallback walkerCallback) {
		super();
		assert null != walkerCallback;
		callback = walkerCallback;
	}
	
	public void processAutomaton(Class<? extends Object> aClass) {
		assert null != aClass;
		processedClasses.clear();
		processedMethods.clear();
		/* TODO examine all states for presence of @Disabled annotation and populate 
		 * disabledStates prior to actual processing the automaton
		 */
		processStates(aClass);
	}
	
	@SuppressWarnings("unchecked")
	private void processStates(Class<? extends Object> aClass) {
		Class<?>[] classes = aClass.getDeclaredClasses();
		for(Class<?> clazz : classes) {
			if( State.class.isAssignableFrom(clazz) ) {
				if( ! processedClasses.contains(clazz) ) {
					processState((Class<? extends State>)clazz);
				}
			}
			else {				
				// Ignoring other classes that can be defined in the automaton
			}
		}
		aClass = aClass.getSuperclass();
		if( (null != aClass) && ! rootClass.equals(aClass) )
			processStates(aClass);
	}

	private void processState(Class<? extends State> clazz){
		processedClasses.add(clazz);
		if( clazz.isAnnotationPresent(Disable.class) ) {
			disabledStates.add(clazz);
			new Traverser<State>(State.class, clazz, processedClasses) {
				@Override public void visit(Class<? extends State> clazz) { disabledStates.add(clazz); } }.traverse();
			return;
		}
		callback.stateBegin(clazz);
		callback.actionsBegin(clazz);
		processStateActions(clazz);
		callback.actionsEnd(clazz);
		callback.transitonsBegin(clazz);
		processTransitions(clazz);
		new Traverser<State>(State.class, clazz, processedClasses) {
			@Override public void visit(Class<? extends State> clazz) { processTransitions(clazz); } }.traverse();
		callback.transitonsEnd(clazz);
		callback.stateEnd(clazz);
	}

	@SuppressWarnings("unchecked")
	private void processTransitions(Class<? extends State> clazz) {
		for(Method method : clazz.getDeclaredMethods()) {
			Class<?> to = method.getReturnType();
			if( ! void.class.equals(to) ) {
				if( State.class.isAssignableFrom(to)) {
					if( ! processedMethods.contains(method) ) {
						processTransition(clazz, (Class<? extends State>) to, method);
					}
				}
				else {
					logError(String.format("Ignored method '%s' '%s' not returning descendants of '%s'", to.getName(), method.getName(), State.class.getCanonicalName()));					
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	private void processTransition(Class<? extends State> from, Class<? extends State> to, Method method) {
		processedMethods.add(method);
		if( method.isAnnotationPresent(Disable.class) ) return;
		if( to.isAnnotationPresent(Disable.class) || disabledStates.contains(to)) {
			logError(String.format("Transition '%s.%s' refers to a disbaled state '%s'",from.getCanonicalName(),method.getName(),to.getSimpleName()));						
		}
		Class<? extends Event> event = findEvent(method);
		if( null == event ) {
			logError(String.format("Event is not found for transition '%s.%s'",from.getCanonicalName(),method.getName()));			
		}		
		callback.transitionBegin(from, to, event, method.getName());
		for(Class<?> param : method.getParameterTypes()) {
			if( Event.class.isAssignableFrom(param) ) {
			} 
			else if( Condition.class.isAssignableFrom(param) ) {
				callback.condition(from, to, event, method.getName(), (Class<? extends Condition>)param);
			}
			else if ( Action.class.isAssignableFrom(param)) {
				callback.action(from, to, event, method.getName(), (Class<? extends Action>) param);
			}
			else {
					logError(String.format("Unexpected parameter type '%s'. Expecting one of  ('%s','%s','%s')", param.getCanonicalName(), Event.class.getName(), Condition.class.getName(), Action.class.getName()));
				}
		}
		callback.transitionEnd(from, to, event, method.getName());
	}

	@SuppressWarnings("unchecked")
	private Class<? extends Event> findEvent(Method method) {
		Class<? extends Event> event = null;
		for(Class<?> param : method.getParameterTypes()) {
			if( Event.class.isAssignableFrom(param) ) {
				if( null == event ) {
					event = (Class<? extends Event>) param;
				}
				else {
					logError(String.format("Ignored parameter '%s' in method '%s': only one event is acceptable",param.getCanonicalName(),method.getName()));
				}
			}
		}
		return event;
	}

	protected void logError(String format) {
		System.err.println(format);
	}

	@SuppressWarnings("unchecked")
	private void processStateActions(Class<? extends State> clazz) {		
		for(Method method : clazz.getDeclaredMethods()) {
			if( processedMethods.contains(method) )
				continue;
			Class<?> to = method.getReturnType();
			if( ! void.class.equals(to) ) 
				continue;
			processedMethods.add(method);
			if( method.isAnnotationPresent(Disable.class) ) {
				
			} 
			else if( method.isAnnotationPresent(OnEnter.class) ) {
				for(Class<?> param : method.getParameterTypes() ) {
					if( Action.class.isAssignableFrom(param)) {
						callback.actionOnEnter(clazz, method.getName(), (Class<? extends Action>) param);
					}
					else {
						logError(String.format("Parameterr '%s' is not '%s'", param.getCanonicalName(), Action.class.getName()));						
					}
				}
			}
			else if( method.isAnnotationPresent(OnLeave.class) ) {					
				for(Class<?> param : method.getParameterTypes() ) {
					if( Action.class.isAssignableFrom(param)) {
						callback.actionOnLeave(clazz, method.getName(), (Class<? extends Action>) param);
					}
					else {
						logError(String.format("Parameterr '%s' is not '%s'", param.getCanonicalName(), Action.class.getName()));						
					}
				}
			}
			else {
				logError(String.format("Method '%s' returning void is missing @OnEnter or @OnLeave",method.getName()));
			}
		}
		new Traverser<State>(State.class, clazz, processedClasses) {
			@Override public void visit(Class<? extends State> clazz) { processStateActions(clazz); } }.traverse();
/*		
		Class<?> superb = clazz.getSuperclass();
		if( null != superb && ! superb.equals(rootClass) && ! processedClasses.contains(superb)) {
			processStateActions((Class<? extends State>) superb);
			processedClasses.add(superb);
		}
*/		
	}
}
